# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from gt4py.next.ffront.decorator import (
    embedded_program_call_context as embedded_program_call_context,
    program_call_context as program_call_context,
)
from gt4py.next.otf.compiled_program import (
    compile_variant_hook as compile_variant_hook,
    compiled_program_call_context as compiled_program_call_context,
)
